from geneal.genetic_algorithms import ContinuousGenAlgSolver, BinaryGenAlgSolver


class TemplateChildClass(ContinuousGenAlgSolver, BinaryGenAlgSolver):
    def __init__(self, *args, **kwargs):
        BinaryGenAlgSolver.__init__(self, *args, **kwargs)
        ContinuousGenAlgSolver.__init__(self, *args, **kwargs)

    def fitness_function(self, chromosome):
        """
        Implements the logic that calculates the fitness
        measure of an individual.

        :param chromosome: chromosome of genes representing an individual
        :return: the fitness of the individual
        """
        pass

    def initialize_population(self, pop_size, n_genes):
        """
        Initializes the population of the problem

        :param pop_size: number of individuals in the population
        :param n_genes: number of genes representing the problem. In case of the binary
        solver, it represents the number of genes times the number of bits per gene
        :return: a numpy array with a randomized initialized population
        """
        pass

    def create_offspring(
        self, first_parent, sec_parent, crossover_pt, offspring_number
    ):
        """
        Creates an offspring from 2 parents. It uses the crossover point(s)
        to determine how to perform the crossover

        :param first_parent: first parent's chromosome
        :param sec_parent: second parent's chromosome
        :param crossover_pt: point(s) at which to perform the crossover
        :param offspring_number: whether it's the first or second offspring from a pair of parents.
        Important if there's different logic to be applied to each case.
        :return: the resulting offspring.
        """
        pass

    def mutate_population(self, population, n_mutations):
        """
        Mutates the population according to a given user defined rule.

        :param population: the population at a given iteration
        :param n_mutations: number of mutations to be performed. This number is
        calculated according to mutation_rate, but can be adjusted as needed inside this function
        :return: the mutated population
        """
        pass



from functools import reduce

import numpy as np
import networkx as nx

from geneal.genetic_algorithms import ContinuousGenAlgSolver


class TravellingSalesmanProblemSolver(ContinuousGenAlgSolver):
    def __init__(self, graph, *args, **kwargs):

        if "n_crossover_points" in kwargs:
            if kwargs["n_crossover_points"] != 2:
                print("Defaulting 'n_crossover_points' to 2")
            kwargs.pop("n_crossover_points")

        ContinuousGenAlgSolver.__init__(self, n_crossover_points=2, *args, **kwargs)

        self.G = graph

    def fitness_function(self, individual):
        """
        Implements the logic that calculates the fitness
        measure of an individual. It sums all the costs of going
        from node to node in the tour.

        :param individual: chromosome of genes representing an individual
        :return: the fitness of the individual
        """

        res = reduce(
            lambda total_length, city_pair: total_length
            + self.G.edges[(city_pair[0], city_pair[1])]["weight"],
            zip(individual, individual[1:]),
            0,
        )

        res += self.G.edges[(241, individual[0])]["weight"]
        res += self.G.edges[(individual[-1], 241)]["weight"]

        return -round(res, 2)

    def initialize_population(self, pop_size, n_genes):
        """
        Initializes the population of the problem. It creates a
        matrix of size (pop_size x n_genes) containing permutations of the nodes
        on each row.

        :param pop_size: number of individuals in the population
        :param n_genes: number of genes representing the problem. In case of the binary
        solver, it represents the number of genes times the number of bits per gene
        :return: a numpy array with a randomized initialized population
        """

        population = np.repeat(
            np.arange(1, n_genes + 1)[np.newaxis, :], pop_size, axis=0
        )

        return np.array(list(map(lambda x: np.random.permutation(x), population)))

    def create_offspring(self, first_parent, sec_parent, crossover_pt, _):
        """
        Creates an offspring from 2 parents. It performs an OX crossover, which
        combines genes from each parent, but maintaining the nodes order of the parents.

        http://www.inf.tu-dresden.de/content/institutes/ki/cl/study/summer14/pssai/slides/GA_for_TSP.pdf

        :param first_parent: first parent's chromosome
        :param sec_parent: second parent's chromosome
        :param crossover_pt: points at which to perform the crossover
        :return: the resulting offspring.
        """

        reordered_sec_parent = np.roll(sec_parent, -crossover_pt[1])

        new_arr = first_parent[crossover_pt[0] : crossover_pt[1]]

        new_arr = np.append(new_arr, reordered_sec_parent)

        _, idx = np.unique(new_arr, return_index=True)

        res = np.roll(new_arr[np.sort(idx)], crossover_pt[0])

        if res.shape[0] != 30:
            a = 1

        return res

    def mutate_population(self, population, n_mutations):
        """
        Mutates the population using a 2-opt rule hybrid. It selects the number of rows
        on which mutation will be applied, and then a applies a local search 2-opt rule
        to those rows.

        :param population: the population at a given iteration
        :param n_mutations: number of mutations to be performed. This number is
        calculated according to mutation_rate, but can be adjusted as needed inside this function
        :return: the mutated population
        """

        adjusted_n_mutations = np.ceil(n_mutations / self.n_genes).astype(int)

        mutation_rows, mutation_cols = self.get_mut_rows_cols(
            adjusted_n_mutations, population
        )

        population[mutation_rows, :] = np.array(
            list(
                map(
                    lambda route: self.local_search(route),
                    population[mutation_rows, :],
                )
            )
        )

        return population

    def local_search(self, route):

        """
        It applies a 2-opt local search to a particular route. It selects 10 combinations
        of nodes on a given tour, performs a 2-opt node flipping,
        calculates the fitness of the resulting new tours and then selects the one
        with maximum fitness

        :param route: a given route to be mutated.
        :return: the mutated route
        """

        mutation_cols = np.sort(np.random.choice(route.shape[0], size=(10, 2)), axis=1)

        local_search_rows = np.array(
            list(
                map(
                    lambda args: self.two_opt_swap(*args),
                    zip(
                        np.repeat(route[np.newaxis, :], mutation_cols.shape[0], axis=0),
                        mutation_cols,
                    ),
                )
            )
        )

        fitness = np.array(
            list(map(lambda row: self.fitness_function(row), local_search_rows))
        )
        max_fitness = np.argmax(fitness)

        return local_search_rows[max_fitness, :]

    @staticmethod
    def two_opt_swap(route, mutation_cols):
        """
        Flips the nodes between 2 node positions for a given route.

        :param route: a given route to be mutated
        :param mutation_cols: the position of the nodes to be flipped in between
        :return: the mutated route
        """
        route[mutation_cols[0] : mutation_cols[1]] = np.flip(
            route[mutation_cols[0] : mutation_cols[1]]
        )

        return route

    @staticmethod
    def get_mut_rows_cols(n_mutations, population):

        mutation_rows = np.random.choice(
            np.arange(1, population.shape[0]), n_mutations, replace=False
        )

        # mutation_cols = np.sort(np.random.choice(population.shape[1], size=(adjusted_n_mutations, 2)), axis=1)

        mutation_cols = np.array(
            list(
                map(
                    lambda x: np.sort(
                        np.random.choice(
                            np.arange(population.shape[1]), 2, replace=False
                        )
                    ),
                    range(n_mutations),
                )
            )
        ).astype(int)

        return mutation_rows, mutation_cols



import numpy as np


def fitness_functions_binary(function_number):

    if function_number == 1:

        return lambda chromosome: chromosome.sum()

    elif function_number == 2:

        answer = np.array(
            [
                1,
                0,
                1,
                1,
                0,
                0,
                0,
                1,
                1,
                1,
                0,
                0,
                1,
                0,
                1,
                1,
                0,
                1,
                1,
                0,
                1,
                0,
                0,
                0,
                1,
                0,
                0,
                1,
                0,
                0,
                1,
                0,
                0,
                0,
                0,
                0,
                1,
                0,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                0,
                0,
                0,
            ]
        )

        return lambda chromosome: -np.abs(answer - chromosome).sum()



from functools import reduce
from itertools import combinations

import numpy as np
from dcor import distance_correlation
from numba import njit, jit
from scipy.stats import pearsonr


association_methods = {
    "DCD": {"association_method": distance_correlation, "relation": "difference"},
    "DCQ": {"association_method": distance_correlation, "relation": "quotient"},
    "PCD": {
        "association_method": lambda x, y: np.abs(pearsonr(x, y)[0]),
        "relation": "difference",
    },
    "PCQ": {
        "association_method": lambda x, y: np.abs(pearsonr(x, y)[0]),
        "relation": "quotient",
    },
}


class AssociationMethods:
    def __init__(self, features, target, method: str = "DCD"):

        if method not in association_methods:
            raise Exception(
                f"Invalid method, must be one of {association_methods}."
                f" Check the documentation for more details"
            )

        if target.ndim != 1:
            raise Exception(f"Invalid target variable, it must be a 1-D array")

        self.features = np.array(features)
        self.target = np.array(target)

        self.association_method = association_methods[method]["association_method"]
        self.relation = association_methods[method]["relation"]

    def fitness_function(self, individual):

        on_features = np.arange(self.features.shape[1])[individual == 1]

        relevance = reduce(
            lambda prev, feature: prev
            + np.nan_to_num(
                self.association_method(self.features[:, feature], self.target)
            ),
            on_features,
            0,
        )

        redundancy = reduce(
            lambda prev, case: prev
            + np.nan_to_num(
                self.association_method(
                    self.features[:, case[0]], self.features[:, case[1]]
                )
            ),
            combinations(on_features, 2),
            0,
        )

        cardinality = on_features.size

        try:
            if self.relation == "difference":
                return relevance / cardinality - redundancy / cardinality ** 2
            else:
                return (relevance / cardinality) / (redundancy / cardinality ** 2)
        except ZeroDivisionError:
            return 0



def get_input_dimensions(lst, n_dim=0):
    if isinstance(lst, (list, tuple)):
        return get_input_dimensions(lst[0], n_dim + 1) if len(lst) > 0 else 0
    else:
        return n_dim


def get_elapsed_time(start_time, end_time):

    runtime = (end_time - start_time).seconds

    hours, remainder = divmod(runtime, 3600)
    minutes, seconds = divmod(remainder, 60)

    time_str = ""

    if hours:
        time_str += f"{hours} hours, "

    if minutes:
        time_str += f"{minutes} minutes, "

    if seconds:
        time_str += f"{seconds} seconds"

    return time_str



from itertools import chain

import numpy as np
from numba import njit, jit


def fitness_function_song(chromosome):
    answer = np.array(
        [
            1,
            0,
            1,
            1,
            0,
            0,
            0,
            1,
            1,
            1,
            0,
            0,
            1,
            0,
            1,
            1,
            0,
            1,
            1,
            0,
            1,
            0,
            0,
            0,
            1,
            0,
            0,
            1,
            0,
            0,
            1,
            0,
            0,
            0,
            0,
            0,
            1,
            0,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            0,
            0,
            0,
        ]
    )

    return -np.abs(answer - chromosome).sum()


def fitness_function_sudoku(chromosome):

    puzzle = chromosome.reshape(9, 9)

    units = np.array(list(chain(*[np.vsplit(sub, 3) for sub in np.hsplit(puzzle, 3)])))

    points = 0

    target = np.arange(1, 10).astype(float)

    for m in range(9):
        points += -0.1 * ((puzzle[m, :]).sum() - 45) ** 2 + 100

        unique = np.unique(puzzle[m, :])

        points += (
            200 if len(unique) == len(target) and np.equal(unique, target).all() else 0
        )

    for n in range(9):
        points += -0.1 * ((puzzle[:, n]).sum() - 45) ** 2 + 100

        unique = np.unique(puzzle[:, n])

        points += (
            200 if len(unique) == len(target) and np.equal(unique, target).all() else 0
        )

    for unit in units:
        points += -0.1 * (unit.sum() - 45) ** 2 + 100

        unique = np.unique(unit.flatten())

        points += (
            200 if len(unique) == len(target) and np.equal(unique, target).all() else 0
        )

    return points


def fitness_function_quadratic(chromosome):

    # if chromosome[0] < -5 or chromosome[0] > 5 or chromosome[1] < -5 or chromosome[1] > 5:
    #     return -np.inf

    return -(chromosome[0] ** 2 + chromosome[1] ** 2)


def fitness_function_2(chromosome):

    if (
        chromosome[0] < 0
        or chromosome[0] > 10
        or chromosome[1] < 0
        or chromosome[1] > 10
    ):
        return 0

    return -(
        chromosome[0] * np.sin(4 * chromosome[0])
        + 1.1 * chromosome[1] * np.sin(2 * chromosome[1])
    )



class NoFitnessFunction(Exception):
    def __init__(self, *args, **kwargs):
        self.message = args[0] if args else "NoFitnessFunction"

    def __str__(self):
        return f"{self.message}"

    def __repr__(self):
        return self.__class__.__name__



import os

import networkx as nx
import pandas as pd
import turf

try:
    module_path = os.path.abspath(os.path.join(".."))
except FileNotFoundError:
    module_path = os.path.abspath(os.path.join("."))

world_capitals = pd.read_csv(os.path.join(module_path, "data/world_capitals.csv"))

world_capitals.dropna(subset=["CapitalName"], axis=0, inplace=True)

world_capitals.reset_index(drop=True, inplace=True)
world_capitals.index += 1

world_capitals_dict = world_capitals.to_dict(orient="index")

G = nx.Graph()

G.add_nodes_from(world_capitals.index)
nx.set_node_attributes(G, world_capitals_dict)

for city, city_attr in world_capitals_dict.items():
    edges = [
        (
            city,
            to_city,
            round(
                turf.distance(
                    [city_attr["CapitalLongitude"], city_attr["CapitalLatitude"]],
                    [to_city_attr["CapitalLongitude"], to_city_attr["CapitalLatitude"]],
                    {"units": "kilometers"},
                ),
                2,
            ),
        )
        for to_city, to_city_attr in world_capitals_dict.items()
        if to_city != city
    ]

    edges = sorted(edges, key=lambda x: x[2])

    G.add_weighted_edges_from(edges)



import numpy as np


def fitness_functions_continuous(function_number):

    if function_number == 1:
        return lambda chromosome: -(np.abs(chromosome[0]) + np.cos(chromosome[0]))
    elif function_number == 2:
        return lambda chromosome: -(np.abs(chromosome[0]) + np.sin(chromosome[0]))
    elif function_number == 3:
        return lambda chromosome: -(chromosome ** 2).sum()
    elif function_number == 4:
        return lambda chromosome: -np.sum(
            np.abs(chromosome) - 10 * np.cos(np.sqrt(np.abs(10 * chromosome)))
        )
    elif function_number == 5:
        return lambda chromosome: -(chromosome[0] ** 2 + chromosome[0]) * np.cos(
            chromosome[0]
        )
    elif function_number == 6:
        return lambda chromosome: -(
            chromosome[0] * np.sin(4 * chromosome[0])
            + 1.1 * chromosome[1] * np.sin(2 * chromosome[1])
        )



import numpy as np

from geneal.genetic_algorithms.genetic_algorithm_base import GenAlgSolver


class BinaryGenAlgSolver(GenAlgSolver):
    def __init__(
        self,
        n_genes: int,
        fitness_function=None,
        n_bits: int = 1,
        max_gen: int = 1000,
        pop_size: int = 100,
        mutation_rate: float = 0.15,
        selection_rate: float = 0.5,
        n_crossover_points: int = 1,
        random_state: int = None,
    ):
        """
        :param fitness_function: can either be a fitness function or a class implementing a fitness function +
        methods to override the default ones: create_offspring, mutate_population, initialize_population
        :param n_genes: number of genes (variables) to have in each chromosome
        :param n_bits: number of bits representing each gene
        :param max_gen: maximum number of generations to perform the optimization
        :param pop_size: population size
        :param mutation_rate: rate at which random mutations occur
        :param selection_rate: percentage of the population to be selected for crossover
        """

        GenAlgSolver.__init__(
            self,
            fitness_function=fitness_function,
            n_genes=n_genes * n_bits,
            max_gen=max_gen,
            pop_size=pop_size,
            mutation_rate=mutation_rate,
            selection_rate=selection_rate,
            n_crossover_points=n_crossover_points,
            random_state=random_state,
        )

    def initialize_population(self, pop_size, n_genes):
        """
        Initializes the population of the problem according to the
        population size and number of genes and outputs individuals
        composed of 0's and 1's.

        :param pop_size: number of individuals in the population
        :param n_genes: number of genes representing the problem, which corresponds to
        the number of genes times the number of bits per gene
        :return: a numpy array with a randomized initialized population
        """
        return np.round(np.random.rand(pop_size, n_genes))

    @staticmethod
    def create_offspring(first_parent, sec_parent, crossover_pt, _):
        """
        Creates an offspring from 2 parents. It performs the crossover
        according the following rule:

        offspring = [first_parent[:crossover_pt], sec_parent[crossover_pt:]
        """

        return np.hstack(
            (first_parent[: crossover_pt[0]], sec_parent[crossover_pt[0] :])
        )

    def mutate_population(self, population, n_mutations):
        """
        Mutates the population by randomizing specific positions of the
        population individuals.

        :param population: the population at a given iteration
        :param n_mutations: number of mutations to be performed.
        :return: the mutated population
        """

        mutation_rows, mutation_cols = super(
            BinaryGenAlgSolver, self
        ).mutate_population(population, n_mutations)

        population[mutation_rows, mutation_cols] = np.abs(population - 1)[
            mutation_rows, mutation_cols
        ]

        return population



import os
from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

try:
    __version__ = os.environ["GITHUB_REF"].split("/")[-1]
    print(f"Version: {__version__}")
except KeyError:
    from geneal.version import __version__

setup(
    name="geneal",
    version=__version__,
    description="Python Genetic Algorithms library",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/diogomatoschaves/geneal",
    author="Diogo Matos Chaves",
    author_email="di.matoschaves@gmail.com",
    packages=[*find_packages(), "geneal.utils"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.5",
    install_requires=["numpy", "matplotlib", "scipy", "networkx", "pyturf", "pandas",],
    test_requires=["pytest", "pytest-cov", "pytest-mock"],
    keywords=["genetic algorithms", "ga", "optimization", "genetic programming"]
)



import numpy as np

from geneal.genetic_algorithms.genetic_algorithm_base import GenAlgSolver
from geneal.utils.helpers import get_input_dimensions


class ContinuousGenAlgSolver(GenAlgSolver):
    def __init__(
        self,
        n_genes: int,
        fitness_function=None,
        max_gen: int = 1000,
        pop_size: int = 100,
        mutation_rate: float = 0.15,
        selection_rate: float = 0.5,
        variables_limits=(-10, 10),
        problem_type=float,
        n_crossover_points: int = 1,
        random_state: int = None,
    ):
        """
        :param fitness_function: can either be a fitness function or
        a class implementing a fitness function + methods to override
        the default ones: create_offspring, mutate_population, initialize_population
        :param n_genes: number of genes (variables) to have in each chromosome
        :param max_gen: maximum number of generations to perform the optimization
        :param pop_size: population size
        :param mutation_rate: rate at which random mutations occur
        :param selection_rate: percentage of the population to be selected for crossover
        :param variables_limits: limits for each variable [(x1_min, x1_max), (x2_min, x2_max), ...].
        If only one tuple is provided, then it is assumed the same for every variable
        :param problem_type: whether problem is of float or integer type
        """

        GenAlgSolver.__init__(
            self,
            fitness_function=fitness_function,
            n_genes=n_genes,
            max_gen=max_gen,
            pop_size=pop_size,
            mutation_rate=mutation_rate,
            selection_rate=selection_rate,
            n_crossover_points=n_crossover_points,
            random_state=random_state,
        )

        if not variables_limits:
            min_max = np.iinfo(np.int64)
            variables_limits = [(min_max.min, min_max.max) for _ in range(n_genes)]

        if get_input_dimensions(variables_limits) == 1:
            variables_limits = [variables_limits for _ in range(n_genes)]

        self.variables_limits = variables_limits
        self.problem_type = problem_type

    def initialize_population(self, pop_size, n_genes):
        """
        Initializes the population of the problem according to the
        population size and number of genes and according to the problem
        type (either integers or floats).

        :param pop_size: number of individuals in the population
        :param n_genes: number of genes representing the problem.
        :return: a numpy array with a randomized initialized population
        """

        population = np.empty(shape=(pop_size, n_genes))

        for i, variable_limits in enumerate(self.variables_limits):
            if self.problem_type == float:
                population[:, i] = np.random.uniform(
                    variable_limits[0], variable_limits[1], size=pop_size
                )
            else:
                population[:, i] = np.random.randint(
                    variable_limits[0], variable_limits[1] + 1, size=pop_size
                )

        return population

    def get_crossover_points(self):
        """
        Retrieves random crossover points

        :return: a numpy array with the crossover points
        """

        return np.sort(
            np.random.choice(
                np.arange(self.n_genes), self.n_crossover_points, replace=False
            )
        )

    def create_offspring(
        self, first_parent, sec_parent, crossover_pt, offspring_number
    ):
        """
        Creates an offspring from 2 parents. It performs the crossover
        according the following rule:

        p_new = first_parent[crossover_pt] + beta * (first_parent[crossover_pt] - sec_parent[crossover_pt])

        offspring = [first_parent[:crossover_pt], p_new, sec_parent[crossover_pt + 1:]

        where beta is a random number between 0 and 1, and can be either positive or negative
        depending on if it's the first or second offspring

        http://index-of.es/z0ro-Repository-3/Genetic-Algorithm/R.L.Haupt,%20S.E.Haupt%20-%20Practical%20Genetic%20Algorithms.pdf

        :param first_parent: first parent's chromosome
        :param sec_parent: second parent's chromosome
        :param crossover_pt: point(s) at which to perform the crossover
        :param offspring_number: whether it's the first or second offspring from a pair of parents.
        Important if there's different logic to be applied to each case.
        :return: the resulting offspring.
        """

        crossover_pt = crossover_pt[0]

        beta = (
            np.random.rand(1)[0]
            if offspring_number == "first"
            else -np.random.rand(1)[0]
        )

        if self.problem_type == float:
            p_new = first_parent[crossover_pt] - beta * (
                first_parent[crossover_pt] - sec_parent[crossover_pt]
            )
        else:
            p_new = first_parent[crossover_pt] - np.round(
                beta * (first_parent[crossover_pt] - sec_parent[crossover_pt])
            )

        return np.hstack(
            (first_parent[:crossover_pt], p_new, sec_parent[crossover_pt + 1 :])
        )

    def mutate_population(self, population, n_mutations):
        """
        Mutates the population by randomizing specific positions of the
        population individuals.

        :param population: the population at a given iteration
        :param n_mutations: number of mutations to be performed.
        :return: the mutated population
        """

        mutation_rows, mutation_cols = super(
            ContinuousGenAlgSolver, self
        ).mutate_population(population, n_mutations)

        population[mutation_rows, mutation_cols] = self.initialize_population(self.pop_size, self.n_genes)[
            mutation_rows, mutation_cols
        ]

        return population



import datetime
import logging
import math
from abc import ABCMeta, abstractmethod

import numpy as np
import matplotlib.pyplot as plt

from geneal.utils.exceptions import NoFitnessFunction
from geneal.utils.helpers import get_elapsed_time
from geneal.utils.logger import configure_logger


class GenAlgSolver:
    def __init__(
        self,
        n_genes: int,
        fitness_function=None,
        max_gen: int = 1000,
        pop_size: int = 100,
        mutation_rate: float = 0.15,
        selection_rate: float = 0.5,
        n_crossover_points: int = 1,
        random_state: int = None,
    ):
        """
        :param fitness_function: can either be a fitness function or
        a class implementing a fitness function + methods to override
        the default ones: create_offspring, mutate_population, initialize_population
        :param n_genes: number of genes (variables) to have in each chromosome
        :param max_gen: maximum number of generations to perform the optimization
        :param pop_size: population size
        :param mutation_rate: rate at which random mutations occur
        :param selection_rate: percentage of the population to be selected for crossover
        """

        if isinstance(random_state, int):
            np.random.seed(random_state)

        configure_logger()

        if not fitness_function:
            try:
                getattr(self, "fitness_function")
            except AttributeError:
                raise NoFitnessFunction(
                    "A fitness function must be defined or provided as an argument"
                )
        else:
            self.fitness_function = fitness_function

        self.n_genes = n_genes
        self.max_gen = max_gen
        self.pop_size = pop_size
        self.mutation_rate = mutation_rate
        self.selection_rate = selection_rate
        self.n_crossover_points = n_crossover_points

        self.pop_keep = math.floor(selection_rate * pop_size)

        self.generations_ = 0
        self.best_fitness_ = 0
        self.best_individual_ = None

    def solve(self):
        """
        Performs the genetic algorithm optimization according to the parameters
        provided at initialization.

        :return: None
        """

        start_time = datetime.datetime.now()

        mean_fitness = np.ndarray(shape=(1, 0))
        max_fitness = np.ndarray(shape=(1, 0))

        # initialize the population
        population = self.initialize_population(self.pop_size, self.n_genes)

        fitness = self.calculate_fitness(population)

        fitness, population = self.sort_by_fitness(fitness, population)

        mating_prob = (
            np.arange(1, self.pop_keep + 1) / np.arange(1, self.pop_keep + 1).sum()
        )[::-1]

        prob_intervals = np.array([0, *np.cumsum(mating_prob[: self.pop_keep + 1])])

        find_parent = np.vectorize(lambda value: np.argmin(value > prob_intervals) - 1)

        number_matings = math.floor((self.pop_size - self.pop_keep) / 2)

        n_mutations = math.ceil((self.pop_size - 1) * self.n_genes * self.mutation_rate)

        gen_interval = round(self.max_gen / 10)

        gen_n = 0
        while True:

            gen_n += 1

            if gen_n % gen_interval == 0:
                logging.info(gen_n)

            mean_fitness = np.append(mean_fitness, fitness.mean())
            max_fitness = np.append(max_fitness, fitness[0])

            ma = find_parent(np.random.rand(number_matings))
            pa = find_parent(np.random.rand(number_matings))

            ix = np.arange(0, self.pop_size - self.pop_keep - 1, 2)  # index of mate #1

            xp = np.array(
                list(map(lambda _: self.get_crossover_points(), range(number_matings)))
            )

            for i in range(xp.shape[0]):

                # create first offspring
                population[self.pop_keep + ix[i], :] = self.create_offspring(
                    population[ma[i], :], population[pa[i], :], xp[i], "first"
                )

                # create second offspring
                population[self.pop_keep + ix[i] + 1, :] = self.create_offspring(
                    population[pa[i], :], population[ma[i], :], xp[i], "second"
                )

            population = self.mutate_population(population, n_mutations)

            fitness = np.hstack((fitness[0], self.calculate_fitness(population[1:, :])))

            fitness, population = self.sort_by_fitness(fitness, population)

            if gen_n >= self.max_gen:
                break

        self.generations_ = gen_n
        self.best_fitness_ = fitness[0]
        self.best_individual_ = population[0, :]
        self.population_ = population

        self.plot_results(mean_fitness, max_fitness, gen_n)

        end_time = datetime.datetime.now()

        time_str = get_elapsed_time(start_time, end_time)

        self.print_stats(time_str)

    def calculate_fitness(self, population):
        """
        Calculates the fitness of the population

        :param population: population state at a given iteration
        :return: the fitness of the current population
        """
        return np.array(list(map(self.fitness_function, population)))

    @staticmethod
    def sort_by_fitness(fitness, population):
        """
        Sorts the population by its fitness.

        :param fitness: fitness of the population
        :param population: population state at a given iteration
        :return: the sorted fitness array and sorted population array
        """

        sorted_fitness = np.argsort(fitness)[::-1]

        population = population[sorted_fitness, :]
        fitness = fitness[sorted_fitness]

        return fitness, population

    def get_crossover_points(self):
        """
        Retrieves random crossover points

        :return: a numpy array with the crossover points
        """
        return np.sort(
            np.random.choice(
                np.arange(self.n_genes + 1), self.n_crossover_points, replace=False
            )
        )

    @staticmethod
    def plot_results(mean_fitness, max_fitness, iterations):
        """
        Plots the evolution of the mean and max fitness of the population

        :param mean_fitness: mean fitness array for each generation
        :param max_fitness: max fitness array for each generation
        :param iterations: total number of generations
        :return: None
        """

        plt.figure(figsize=(7, 7))

        x = np.arange(1, iterations + 1)

        plt.plot(x, mean_fitness, label="mean fitness")
        plt.plot(x, max_fitness, label="max fitness")

        plt.legend()
        plt.show()

    def print_stats(self, time_str):
        """
        Prints the statistics of the optimization run

        :param time_str: time string given by the method get_elapsed_time
        :return: None
        """

        logging.info("\n#############################")
        logging.info("#\t\t\tSTATS\t\t\t#")
        logging.info("#############################\n\n")
        logging.info(f"Total running time: {time_str}\n\n")
        logging.info(f"Population size: {self.pop_size}")
        logging.info(f"Number variables: {self.n_genes}")
        logging.info(f"Selection rate: {self.selection_rate}")
        logging.info(f"Mutation rate: {self.mutation_rate}")
        logging.info(f"Number Generations: {self.generations_}\n")
        logging.info(f"Best fitness: {self.best_fitness_}")
        logging.info(f"Best individual: {self.best_individual_}")

    @abstractmethod
    def initialize_population(self, pop_size, n_genes):
        """
        Initializes the population of the problem. To be implemented in each child class.

        :param pop_size: number of individuals in the population
        :param n_genes: number of genes representing the problem. In case of the binary
        solver, it represents the number of genes times the number of bits per gene
        :return: a numpy array with a randomized initialized population
        """
        pass

    @staticmethod
    @abstractmethod
    def create_offspring(first_parent, sec_parent, crossover_pt, offspring_number):
        """
        Creates an offspring from 2 parents. It uses the crossover point(s)
        to determine how to perform the crossover. To be implemented on each child class.

        :param first_parent: first parent's chromosome
        :param sec_parent: second parent's chromosome
        :param crossover_pt: point(s) at which to perform the crossover
        :param offspring_number: whether it's the first or second offspring from a pair of parents.
        Important if there's different logic to be applied to each case.
        :return: the resulting offspring.
        """
        pass

    def mutate_population(self, population, n_mutations):
        """
        Mutates the population according to a given user defined rule.
        To be defined further in each child class. Each direct child class can call
        this super method to retrieve the mutation rows and mutations columns

        :param population: the population at a given iteration
        :param n_mutations: number of mutations to be performed. This number is
        calculated according to mutation_rate, but can be adjusted as needed inside this function
        :return: an array with the mutation_rows and mutation_cols
        """

        mutation_rows = np.ceil(
            np.random.rand(1, n_mutations) * (self.pop_size - 1)
        ).astype(int)

        mutation_cols = (
            np.ceil(np.random.rand(1, n_mutations) * self.n_genes) - 1
        ).astype(int)

        return [mutation_rows, mutation_cols]
